package com.syjy.calculate.service;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.googlecode.aviator.AviatorEvaluator;
import com.syjy.calculate.entity.tCalculationFormula;
import com.syjy.calculate.repository.CalculationFormulaRepository;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 计算准确率
 *
 * @author zcl
 * @version 1.0
 * @since 2022/8/8 9:30
 */
@Service
@Slf4j
public class AccuracyPassRateCalculateService {
    BigDecimal ZERO = new BigDecimal("0");
    BigDecimal HUNDRED = new BigDecimal("100");
    String PASS = "PASS";
    @Autowired
    private CalculationFormulaRepository calculationFormulaRepository;

    /**
     * @param powerData    功率相关数据 Map中必要字段：sj(实际功率)、yc(预测功率)
     * @param rl           容量
     * @param provinceEnum 省调
     * @param type         短期、超短期
     * @param compare      通过率对比系数
     * @return 返回Json  result:fail/success  value: 99%  msg:失败及原因
     */
    public JSONObject calculate(JSONArray powerData, BigDecimal rl, String provinceEnum, String type, BigDecimal compare) {
        JSONObject jsonResult = new JSONObject();
        jsonResult.put("result", false);
        // 获取公式
        List<tCalculationFormula> calculationFormulaList = getCalculationFormulaData(type, provinceEnum);
        if (calculationFormulaList == null || calculationFormulaList.size() == 0) {
            jsonResult.put("msg", "计算失败，未匹配到公式");
            return jsonResult;
        }
        // 获取功率个数
        BigDecimal count = BigDecimal.valueOf(powerData.size());
        // 存放每步计算结果
        Map<String, Object> resultMap = new HashMap<>();
        // 公式
        String formula;
        // 循环公式，依次执行公式
        for (tCalculationFormula calculationFormula : calculationFormulaList) {
            BigDecimal formulaResult = ZERO;
            // 获取公式
            formula = calculationFormula.getFormula();
            // 如果包含求和，则进行循环求和计算
            if (formula.contains("sum")) {
                // 获取sum后面的公式
                formula = formula.split(":")[1];
                // 循环执行sum后的公式
                for (int i = 0; i < powerData.size(); i++) {
                    JSONObject dataJson = powerData.getJSONObject(i);
                    // 将执行过的公式结果放入dataMap
                    dataJson.putAll(resultMap);
                    dataJson.put("rl", rl);
                    dataJson.put("count", count);
                    try {
                        BigDecimal thisResult = new BigDecimal(String.valueOf(AviatorEvaluator.execute(formula, dataJson)));
                        // 如果是通过率计算
                        if (this.PASS.equals(type)) {
                            // 如果对比系数是0，返回计算失败
                            if (compare.compareTo(this.ZERO) == 0) {
                                jsonResult.put("msg", "计算失败，通过率计算，对比系数不可为0");
                                return jsonResult;
                            }
                            // 计算结果>=对比系数，则为100，否则为0
                            if (thisResult.compareTo(compare) > -1) {
                                thisResult = this.HUNDRED;
                            } else {
                                thisResult = this.ZERO;
                            }
                        }
                        // 求和计算 AviatorEvaluator.execute(公式，公式中的参数及值(以Map形式存放))
                        formulaResult = formulaResult.add(thisResult);
                    } catch (Exception e) {
                        e.printStackTrace();
                        log.error("计算公式:" + formula + "错误:" + e.toString());
                        jsonResult.put("msg", "计算失败" + "错误:" + e.toString());
                        return jsonResult;
                    }
                }
                // 非求和计算
            } else {
                if (powerData.size() > 0) {
                    // 获取数据测试
                    JSONObject dataJson = powerData.getJSONObject(0);
                    // 将执行过的公式结果放入dataMap
                    dataJson.putAll(resultMap);
                    dataJson.put("rl", rl);
                    dataJson.put("count", count);
                    try {
                        // 根据公式进行计算
                        formulaResult = new BigDecimal(String.valueOf(AviatorEvaluator.execute(formula, dataJson)));
                    } catch (Exception e) {
                        log.error("计算公式:" + formula + "错误:" + e.toString());
                        e.printStackTrace();
                        jsonResult.put("msg", "计算失败" + "错误:" + e.toString());
                        return jsonResult;
                    }
                }
            }
            // 记录本次公式的计算结果
            resultMap.put("result" + calculationFormula.getOrderNo(), formulaResult);
            log.info("result" + calculationFormula.getOrderNo() + ":" + formulaResult);
        }
        // 获取最后一个公式计算的结果
        String lastResult = String.valueOf(resultMap.get("result" + calculationFormulaList.size()));
        if (lastResult == null) {
            jsonResult.put("msg", "计算失败，结果为空");
            return jsonResult;
        }
        log.info("最后一个公式计算的结果:" + lastResult);
        BigDecimal bResult = new BigDecimal(lastResult);
        bResult = bResult.setScale(2, BigDecimal.ROUND_HALF_UP);
        // 过滤准确率
        bResult = filterResult(bResult);
        lastResult = bResult + "%";
        jsonResult.put("result", true);
        jsonResult.put("value", lastResult);
        log.info("准确率:" + lastResult);
        return jsonResult;
    }

    /**
     * 获取当前省调的公式列表
     *
     * @param type         类型
     * @param provinceEnum 省调(在数据库中省调存储必须带分号，例如:E12;)
     * @return
     */
    public List<tCalculationFormula> getCalculationFormulaData(String type, String provinceEnum) {
        List<tCalculationFormula> calculationFormulaList = new ArrayList<>();
        // 根据类型从数据库中查出公式列表
        calculationFormulaList = calculationFormulaRepository.findByTypeAndProvince(type, provinceEnum);
        return calculationFormulaList;
    }

    /**
     * 过滤结果 <0=0  >100=100
     *
     * @param result 过滤前的结果
     * @return 过滤后的结果
     */
    private BigDecimal filterResult(BigDecimal result) {
        //当结果为负数时，说明偏差过大，准确率为0
        if (result.compareTo(BigDecimal.ZERO) == -1 || result.compareTo(BigDecimal.ZERO) == 0) {
            result = BigDecimal.ZERO;
        }

        if (result.compareTo(this.HUNDRED) == 1) {
            result = this.HUNDRED;
        }
        return result;
    }




}

